"""Class to work with Kepler data PRF"""

from typing import Tuple, List
import numpy as np
from .utils import channel_to_module_output, LKPRFWarning
from .data import get_kepler_prf_file
import warnings

from .prfmodel import PRF


class KeplerPRF(PRF):
    """
    A KeplerPRF class. Can be used for Kepler or K2.

    There are 5 PRF measurements (the 4 corners and the center) for each channel.
    The measured PRF is over-sampled by a factor of 50 to enable for sub-pixel interpolation.
    The model is a 550x550 (or 750x750) grid that covers 11x11 (or 15x15) pixels

    https://archive.stsci.edu/missions/kepler/commissioning_prfs/
    """

    def __init__(self, channel: int):
        super().__init__()
        self.channel = channel
        self.mission = "Kepler"
        self._prepare_prf()

    def __repr__(self):
        return f"KeplerPRF Object [Channel {self.channel}]"

    def _get_prf_data(self):
        module, output = channel_to_module_output(self.channel)
        return get_kepler_prf_file(module=module, output=output)

    def update_coordinates(self, targets: List[Tuple], shape: Tuple):
        row, column = self._unpack_targets(targets)
        """Check that coordinates are within bounds, raise warnings otherwise."""
        if (np.atleast_1d(column) < 12).any():
            warnings.warn(
                "`targets` contains collateral pixels: Column(s) < 12",
                LKPRFWarning,
            )

        if ((np.atleast_1d(column) + shape[1]) > 1112).any():
            warnings.warn(
                "`targets` contains collateral pixels: Column(s) > 1112 ",
                LKPRFWarning,
            )
        if (np.atleast_1d(row) < 20).any():
            warnings.warn(
                "`targets` contains collateral pixels: Row(s) < 20",
                LKPRFWarning,
            )
        if ((np.atleast_1d(row) + shape[0]) > 1044).any():
            warnings.warn(
                "`targets` contains collateral pixels: Row(s) > 1044)",
                LKPRFWarning,
            )
        self._update_coordinates(targets=targets, shape=shape)
        return



# Standard library
import os  # noqa
import logging

PACKAGEDIR = os.path.abspath(os.path.dirname(__file__))

version = "1.0.0"
logger = logging.getLogger("lkprf")

from .data import *  # noqa
from .keplerprf import KeplerPRF  # noqa
from .tessprf import TESSPRF  # noqa



"""PRF base class"""

from abc import ABC, abstractmethod
from typing import Tuple, List
import numpy.typing as npt
import numpy as np
from scipy.interpolate import RectBivariateSpline


class PRF(ABC):
    @abstractmethod
    def __init__(self):
        """
        A generic base class object for PRFs. No to be used directly.
        See KeplerPRF and TESSPRF for the instantiable classes.

        Parameters:
        -----------
        column : int
                pixel coordinate of the lower left column value
        row : int
                pixel coordinate of the lower left row value
        shape : tuple
                shape of the resultant PRFs in pixels
        """

    def __repr__(self):
        return "PRF Base Class"

    def __call__(
        self,
        targets: List[Tuple] = [(5.5, 5.5)],
        shape: Tuple = (11, 11),
        origin: Tuple = (0, 0),
    ) -> npt.ArrayLike:
        return self.evaluate(targets=targets, shape=shape, origin=origin)

    def _unpack_targets(self, targets):
        """Take input targets and convert them to row/column arrays"""
        if isinstance(targets, tuple):
            targets = [targets]
        if not isinstance(targets, list):
            raise ValueError("Input a list of targets.")
        if not isinstance(targets[0], tuple):
            raise ValueError(
                "Input targets as tuples with format (row position, column position)."
            )

        target_row, target_column = np.asarray(targets).T
        return target_row, target_column

    def _evaluate(
        self,
        targets: List[Tuple] = [(5.5, 5.5)],
        origin: Tuple = (0, 0),
        shape: Tuple = (11, 11),
        dx=0,
        dy=0,
    ):
        """
        Interpolates the PRF model onto detector coordinates. Hidden function

        Parameters
        ----------
        targets : List of Tuples
            Coordinates of the targets
        origin : Tuple
            The origin of the image, combined with shape this sets the extent of the image
        shape : Tuple
            The shape of the image, combined with the origin this sets the extent of the image

        Returns
        -------
        prf : 3D array
            Three dimensional array representing the PRF values parametrized by flux and centroids.
            Has shape (ntargets, shape[0], shape[1])
        """
        self.update_coordinates(targets=targets, shape=shape)
        target_row, target_column = self._unpack_targets(targets)

        # Integer extent from the PRF model
        r1, r2 = int(np.floor(self.PRFrow[0])), int(np.ceil(self.PRFrow[-1]))
        c1, c2 = int(np.floor(self.PRFcol[0])), int(np.ceil(self.PRFcol[-1]))
        # Position in the PRF model for each source position % 1
        delta_row, delta_col = (
            np.arange(r1, r2)[:, None] + 0.5 - np.atleast_1d(target_row) % 1,
            np.arange(c1, c2)[:, None] + 0.5 - np.atleast_1d(target_column) % 1,
        )

        # prf model for each source, downsampled to pixel grid
        prf = np.asarray(
            [
                self.interpolate(dr, dc, dx=dx, dy=dy)
                for dr, dc in zip(delta_row.T, delta_col.T)
            ]
        )
        # Normalize to ensure no flux loss
        prf /= prf.sum(axis=(1, 2))[:, None, None]

        # Insert values into final array of the correct shape
        ar = np.zeros((len(target_column), *shape))
        for idx, r, c, p in zip(range(len(prf)), target_row, target_column, prf):
            # pixel offset for source
            roffset = int(r - r % 1)
            coffset = int(c - c % 1)
            # row and column position for source
            R, C = np.arange(r1 + roffset, r2 + roffset), np.arange(
                c1 + coffset, c2 + coffset
            )
            # check if pixels are in the resultant image
            k = (R >= origin[0]) & (R < (origin[0] + shape[0]))
            j = (C >= origin[1]) & (C < (origin[1] + shape[1]))
            if k.any() & j.any():
                # if yes, insert into the resultant image
                X, Y = np.meshgrid(R[k] - origin[0], C[j] - origin[1], indexing="ij")
                ar[idx, X, Y] = p[k][:, j]
        return ar

    def evaluate(
        self,
        targets: List[Tuple] = [(5.5, 5.5)],
        origin: Tuple = (0, 0),
        shape: Tuple = (11, 11),
    ):
        """
        Interpolates the PRF model onto detector coordinates.

        Parameters
        ----------
        targets : List of Tuples
            Coordinates of the targets
        origin : Tuple
            The origin of the image, combined with shape this sets the extent of the image
        shape : Tuple
            The shape of the image, combined with the origin this sets the extent of the image

        Returns
        -------
        prf : 3D array
            Three dimensional array representing the PRF values parametrized by flux and centroids.
            Has shape (ntargets, shape[0], shape[1])
        """
        self.update_coordinates(targets=targets, shape=shape)
        return self._evaluate(targets=targets, shape=shape, origin=origin, dx=0, dy=0)

    def gradient(
        self,
        targets: List[Tuple] = [(5.5, 5.5)],
        origin: Tuple = (0, 0),
        shape: Tuple = (11, 11),
    ) -> Tuple:
        """
        Interpolates the gradient of the PRF model onto detector coordinates.

        Parameters
        ----------
        targets : List of Tuples
            Coordinates of the targets
        origin : Tuple
            The origin of the image, combined with shape this sets the extent of the image
        shape : Tuple
            The shape of the image, combined with the origin this sets the extent of the image

        Returns
        -------
        deriv_row, deriv_col : Tuple of two 3D arrays
            This tuple contains two 3D arrays representing the gradient of the PRF values parametrized by flux and centroids.
            Returns (gradient in row, gradient in column). Each array has shape (ntargets, shape[0], shape[1])
        """
        self.update_coordinates(targets=targets, shape=shape)
        deriv_col = self._evaluate(
            targets=targets, shape=shape, origin=origin, dx=0, dy=1
        )
        deriv_row = self._evaluate(
            targets=targets, shape=shape, origin=origin, dx=1, dy=0
        )
        return deriv_row, deriv_col

    @abstractmethod
    def _get_prf_data(self):
        """Method to open PRF files for given mission"""
        pass

    def _prepare_prf(self):
        """
        Sets up the PRF model interpolation by reading in the relevant files,
        and combining them by weighting them by distance to the location on the CCD of interest
        """

        hdulist = self._get_prf_data()
        PRFdata, crval1p, crval2p, cdelt1p, cdelt2p = [], [], [], [], []
        for hdu in hdulist[1:]:
            PRFdata.append(hdu.read())
            hdr = hdu.read_header()
            crval1p.append(hdr["CRVAL1P"])
            crval2p.append(hdr["CRVAL2P"])
            cdelt1p.append(hdr["CDELT1P"])
            cdelt2p.append(hdr["CDELT2P"])

        PRFdata, crval1p, crval2p, cdelt1p, cdelt2p = (
            np.asarray(PRFdata),
            np.asarray(crval1p),
            np.asarray(crval2p),
            np.asarray(cdelt1p),
            np.asarray(cdelt2p),
        )
        PRFdata /= PRFdata.sum(axis=(1, 2))[:, None, None]
        PRFcol = np.arange(0.5, np.shape(PRFdata[0])[1] + 0.5)
        PRFrow = np.arange(0.5, np.shape(PRFdata[0])[0] + 0.5)

        PRFcol = np.arange(0, np.shape(PRFdata[0])[1] + 0)
        PRFrow = np.arange(0, np.shape(PRFdata[0])[0] + 0)

        # Shifts pixels so it is in pixel units centered on 0
        PRFcol = (PRFcol - np.size(PRFcol) / 2) * cdelt1p[0]
        PRFrow = (PRFrow - np.size(PRFrow) / 2) * cdelt2p[0]
        (
            self.PRFrow,
            self.PRFcol,
            self.PRFdata,
            self.crval1p,
            self.crval2p,
            self.cdelt1p,
            self.cdelt2p,
        ) = (PRFrow, PRFcol, PRFdata, crval1p, crval2p, cdelt1p, cdelt2p)

    def _update_coordinates(self, targets: List[Tuple], shape: Tuple):
        row, column = self._unpack_targets(targets)
        # Set the row and column for the model
        row, column = np.atleast_1d(row), np.atleast_1d(column)
        row = np.min([np.max([row, row**0 * 20], axis=0), row**0 * 1044], axis=0).mean()
        column = np.min(
            [np.max([column, column**0 * 12], axis=0), column**0 * 1112], axis=0
        ).mean()

        # interpolate the calibrated PRF shape to the target position
        min_prf_weight = 1e-6
        rowdim, coldim = shape[0], shape[1]
        ref_column = column + 0.5 * coldim
        ref_row = row + 0.5 * rowdim
        supersamp_prf = np.zeros(self.PRFdata.shape[1:], dtype="float32")

        for i in range(self.PRFdata.shape[0]):
            prf_weight = np.sqrt(
                (ref_column - self.crval1p[i]) ** 2 + (ref_row - self.crval2p[i]) ** 2
            )
            if prf_weight < min_prf_weight:
                prf_weight = min_prf_weight
            supersamp_prf += self.PRFdata[i] / prf_weight

        supersamp_prf /= np.nansum(supersamp_prf) * self.cdelt1p[0] * self.cdelt2p[0]
        self.interpolate = RectBivariateSpline(self.PRFrow, self.PRFcol, supersamp_prf)
        return

    @abstractmethod
    def update_coordinates(self, targets, shape):
        """Method to update the interpolation function

        Wrap this parent method, use the public method to check that e.g. targets are in bounds.
        This method should provide the interpolation function
        """
        pass



"""Module to get the archived data for PRFs"""

import os  # noqa
import requests
import numpy as np
import fitsio
from scipy.ndimage import label, uniform_filter

from . import logger, PACKAGEDIR

__all__ = [
    "download_kepler_prf_file",
    "get_kepler_prf_file",
    "build_tess_prf_file",
    "get_tess_prf_file",
    "clear_kepler_cache",
    "clear_tess_cache",
]


_tess_prefixes = {
    1: {
        1: "2019107181900",
        2: "2019107181901",
        3: "2019107181901",
        4: "2019107181901",
    },
    2: {
        1: "2019107181901",
        2: "2019107181901",
        3: "2019107181901",
        4: "2019107181901",
    },
    3: {
        1: "2019107181901",
        2: "2019107181902",
        3: "2019107181902",
        4: "2019107181902",
    },
    4: {
        1: "2019107181902",
        2: "2019107181902",
        3: "2019107181902",
        4: "2019107181902",
    },
}


def _download_file(url, file_path):
    """
    Download the required data files to the specified directory.
    """
    if not os.path.exists(file_path):
        try:
            os.makedirs(os.path.dirname(file_path), exist_ok=True)
            response = requests.get(url)
            response.raise_for_status()  # Ensure the request was successful
            with open(file_path, "wb") as f:
                f.write(response.content)
        except requests.exceptions.HTTPError as http_err:
            if response.status_code == 404:
                return
            else:
                raise http_err


def download_kepler_prf_file(module: int, output: int):
    """Download a Kepler Module file"""
    filename = f"kplr{module:02}.{output}_2011265_prf.fits"
    file_path = f"{PACKAGEDIR}/data/{filename}"
    url = f"https://archive.stsci.edu/missions/kepler/fpc/prf/{filename}"
    logger.info(f"Downloading {module:02}.{output}")
    _download_file(url, file_path)
    hdulist = fitsio.FITS(file_path)
    for hdu in hdulist:
        hdu.verify_checksum()
    return


def build_tess_prf_file(camera: int, ccd: int):
    """Download a set of TESS PRF files for a given camera/ccd"""

    def open_file(url: str):
        """Open a TESS PRF file and correct the edges to 0"""
        file_name = url.split('/')[-1]
        _download_file(url, f"/tmp/{file_name}")
        hdulist = fitsio.FITS(f"/tmp/{file_name}")
        data = hdulist[0].read()
        hdr = hdulist[0].read_header()
        aper = data > 0.0005  # cutout_model > np.percentile(prf, 50)
        labels = label(aper)[0]  # assign group labels to the image
        psf_region = np.bincount(
            labels[aper].flatten()
        ).argmax()  # find the biggest region that contains the prf core
        mask = uniform_filter(
            labels != psf_region, size=10
        )  # apply a filter to smooth out the region
        data[mask] = 0
        return data, hdr

    tess_archive_url = (
        "https://archive.stsci.edu/missions/tess/models/prf_fitsfiles/start_s0004/"
    )
    prefix = _tess_prefixes[camera][ccd]
    filename = f"tess-prf-{camera}-{ccd}.fits"
    file_path = f"{PACKAGEDIR}/data/{filename}"

    R, C = np.meshgrid([1, 513, 1025, 1536, 2048], [45, 557, 1069, 1580, 2092])
    if os.path.isfile(file_path):
        os.remove(file_path)

    hdulist = fitsio.FITS(file_path, mode="rw")
    for (
        r,
        c,
    ) in zip(R.ravel(), C.ravel()):
        url = f"{tess_archive_url}cam{camera}_ccd{ccd}/tess{prefix}-prf-{camera}-{ccd}-row{r:04}-col{c:04}.fits"
        logger.info(f"Downloading CCD {ccd}, Camera {camera}, row {r}, column {c}")
        data, hdr = open_file(url=url)
        hdulist.write(data, header=hdr)
    hdulist.close()
    return


def get_kepler_prf_file(module: int, output: int):
    """Download a Kepler Module file"""
    filename = f"kplr{module:02}.{output}_2011265_prf.fits"
    file_path = f"{PACKAGEDIR}/data/{filename}"
    if not os.path.isfile(file_path):
        logger.info(
            f"No local files found, building Kepler PRF for Module {module}, output {output}."
        )
        download_kepler_prf_file(module=module, output=output)
    file_path = f"{PACKAGEDIR}/data/{filename}"
    hdulist = fitsio.FITS(file_path)
    return hdulist


def get_tess_prf_file(camera: int, ccd: int):
    """Get a PRF file for a given camera/ccd"""
    filename = f"tess-prf-{camera}-{ccd}.fits"
    file_path = f"{PACKAGEDIR}/data/{filename}"
    if not os.path.isfile(file_path):
        logger.info(
            f"No local files found, building TESS PRF for Camera {camera}, CCD {ccd}."
        )
        build_tess_prf_file(camera=camera, ccd=ccd)
    hdulist = fitsio.FITS(file_path)
    return hdulist


def clear_kepler_cache():
    for module in np.arange(26):
        for output in np.arange(1, 5):
            filename = f"kplr{module:02}.{output}_2011265_prf.fits"
            file_path = f"{PACKAGEDIR}/data/{filename}"
            if os.path.isfile(file_path):
                os.remove(file_path)


def clear_tess_cache():
    for camera in np.arange(1, 5):
        for ccd in np.arange(1, 5):
            filename = f"tess-prf-{camera}-{ccd}.fits"
            file_path = f"{PACKAGEDIR}/data/{filename}"
            if os.path.isfile(file_path):
                os.remove(file_path)



"""Class to work with TESS data PRF"""

from typing import Tuple, List
import numpy as np
from .utils import LKPRFWarning
from .data import get_tess_prf_file
import warnings

from .prfmodel import PRF


class TESSPRF(PRF):
    """A TESSPRF class. The TESS PRF measurements are supersampled by a factor of 9."""

    def __init__(self, camera: int, ccd: int):
        super().__init__()
        self.camera = camera
        self.ccd = ccd
        self.mission = "TESS"
        self._prepare_prf()

    def __repr__(self):
        return f"TESSPRF Object [Camera {self.camera}, CCD {self.ccd}]"

    def _get_prf_data(self):
        return get_tess_prf_file(camera=self.camera, ccd=self.ccd)

    def update_coordinates(self, targets: List[Tuple], shape: Tuple):
        row, column = self._unpack_targets(targets)
        """Check that coordinates are within bounds, raise warnings otherwise."""
        if (np.atleast_1d(column) < 45).any():
            warnings.warn(
                "`targets` contains collateral pixels: Column(s) < 45",
                LKPRFWarning,
            )

        if ((np.atleast_1d(column) + shape[1]) > 2093).any():
            warnings.warn(
                "`targets` contains collateral pixels: Column(s) > 2093 ",
                LKPRFWarning,
            )
        if (np.atleast_1d(row) < 0).any():
            warnings.warn(
                "`targets` contains collateral pixels: Row(s) < 0",
                LKPRFWarning,
            )
        if ((np.atleast_1d(row) + shape[0]) > 2048).any():
            warnings.warn(
                "`targets` contains collateral pixels: Row(s) > 2048)",
                LKPRFWarning,
            )
        self._update_coordinates(targets=targets, shape=shape)
        return
